/*
  SDL_Audiolib - An audio mixing, decoding and resampling library
  Copyright (C) 2013  Nikos Chantziaras

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/
#include "audioresampler.h"

#include <algorithm>
#include <cstring>
#include <cmath>
#include "am_util.h"
#include "audiodecoder.h"

using namespace std;
using Aulib::AudioResampler;

#if 0
static unsigned
findGcd(unsigned a, unsigned b)
{
    int remainder;

    while (a) {
        remainder = b % a;
        b = a;
        a = remainder;
    }
    return b ? b : 1;
}
#endif


class Aulib::AudioResampler_priv {
    friend class AudioResampler;
    AudioResampler* q;

    AudioResampler_priv(AudioResampler *pub);
    ~AudioResampler_priv();

    AudioDecoder* fDecoder;
    unsigned fDstRate;
    unsigned fSrcRate;
    unsigned fChannels;
    unsigned fChunkSize;
    float* fOutBuffer;
    float* fInBuffer;
    int fOutBufferSize;
    int fOutBufferPos;
    int fOutBufferEnd;
    int fInBufferSize;
    int fInBufferPos;
    int fInBufferEnd;
    bool fPendingSpecChange;

    int fMoveFromOutBuffer(float dst[], int dstLen);
    int fRelocateInBuffer();
    int fRelocateOutBuffer();
    void fAdjustBufferSizes();
    int fResampleFromInBuffer();
};


Aulib::AudioResampler_priv::AudioResampler_priv(AudioResampler* pub)
    : q(pub),
      fDecoder(nullptr),
      fDstRate(0),
      fSrcRate(0),
      fChannels(0),
      fChunkSize(0),
      fOutBuffer(nullptr),
      fInBuffer(nullptr),
      fOutBufferSize(0),
      fOutBufferPos(0),
      fOutBufferEnd(0),
      fInBufferSize(0),
      fInBufferPos(0),
      fInBufferEnd(0),
      fPendingSpecChange(false)
{
}


Aulib::AudioResampler_priv::~AudioResampler_priv()
{
    delete[] fInBuffer;
    delete[] fOutBuffer;
}


int
Aulib::AudioResampler_priv::fMoveFromOutBuffer(float dst[], int dstLen)
{
    if (fOutBufferEnd <= 0) {
        return 0;
    }
    if (fOutBufferPos >= fOutBufferEnd) {
        fOutBufferPos = fOutBufferEnd = 0;
        return 0;
    }
    int len = std::min(fOutBufferEnd - fOutBufferPos, dstLen);
    memcpy(dst, fOutBuffer + fOutBufferPos, len * sizeof(*fOutBuffer));
    fOutBufferPos += len;
    if (fOutBufferPos >= fOutBufferEnd) {
        fOutBufferEnd = fOutBufferPos = 0;
    }
    //cerr << "moved from outbuf: " << len << endl;
    return len;
}


int
Aulib::AudioResampler_priv::fRelocateInBuffer()
{
    if (fInBufferEnd <= 0) {
        return fInBufferSize;
    }
    if (fInBufferPos >= fInBufferEnd) {
        fInBufferPos = fInBufferEnd = 0;
        return fInBufferSize;
    }
    if (fInBufferPos == 0) {
        return 0;
    }
    int len = fInBufferEnd - fInBufferPos;
    AM_debugPrintLn("relocating input buffer: " << len);
    AM_debugPrintLn("inPos: " << fInBufferPos << " inEnd: " << fInBufferEnd);
    memmove(fInBuffer, fInBuffer + fInBufferPos, len * sizeof(*fInBuffer));
    fInBufferPos = 0;
    fInBufferEnd = len;
    return fInBufferSize - len;
}


int
Aulib::AudioResampler_priv::fRelocateOutBuffer()
{
    if (fOutBufferEnd <= 0)
        return fOutBufferSize;

    if (fOutBufferPos >= fOutBufferEnd) {
        fOutBufferPos = fOutBufferEnd = 0;
        return fOutBufferSize;
    }

    int len = fOutBufferEnd - fOutBufferPos;
    //std::cerr << "relocating output buffer: " << len << '\n';
    memmove(fOutBuffer, fOutBuffer + fOutBufferPos, len * sizeof(*fOutBuffer));
    fOutBufferPos = 0;
    fOutBufferEnd = len;
    return fOutBufferSize - len;
}


void
Aulib::AudioResampler_priv::fAdjustBufferSizes()
{
    int oldInBufLen = fInBufferEnd - fInBufferPos;
    float* oldInBufData = nullptr;
    if (oldInBufLen > 0) {
        AM_debugPrintLn("old data: " << oldInBufLen);
        oldInBufData = new float[oldInBufLen];
        std::memcpy(oldInBufData, fInBuffer + fInBufferPos,
                    oldInBufLen * sizeof(*fInBuffer));
    }

    delete[] fOutBuffer;
    delete[] fInBuffer;

    int minBufSize = fChannels * fChunkSize;
    if (fDstRate == fSrcRate) {
        fOutBufferSize = fInBufferSize = minBufSize;
    } else {
#if 0
        int gcd = findGcd(fDstRate, fSrcRate);
        int ratio_den = fDstRate / gcd;
        int ratio_num = fSrcRate / gcd;
        std::cerr << "CHUNK: " << fChunkSize << '\n';
        fOutBufferSize = minBufSize + (ratio_den % minBufSize) % fChannels;
        fInBufferSize = ratio_num * (fOutBufferSize / ratio_den);
        std::cerr << "MOD: " << (ratio_den % minBufSize) << '\n';
        std::cerr << "den: " << ratio_den << " num: " << ratio_num << '\n';
#else
        fOutBufferSize = minBufSize;
        fInBufferSize = std::ceil((double)fOutBufferSize * fSrcRate / fDstRate);
        int rmd = fInBufferSize % fChannels;
        if (rmd) {
            fInBufferSize = fInBufferSize + fChannels - rmd;
        }
#endif
    }
    AM_debugPrintLn("outbuf: " << fOutBufferSize << " inbuf: " << fInBufferSize);
    fOutBuffer = new float[fOutBufferSize];
    fInBuffer = new float[fInBufferSize];

    fOutBufferPos = fOutBufferEnd = 0;
    fInBufferPos = 0;
    if (oldInBufData) {
        std::memcpy(fInBuffer, oldInBufData, oldInBufLen * sizeof(*oldInBufData));
        fInBufferEnd = oldInBufLen;
    }
    delete[] oldInBufData;
}


int
Aulib::AudioResampler_priv::fResampleFromInBuffer()
{
    int outLen = fOutBufferSize - fOutBufferEnd;
    int inLen = fInBufferEnd - fInBufferPos;
    float* from = fInBuffer + fInBufferPos;
    float* to = fOutBuffer + fOutBufferEnd;
    if (fSrcRate == fDstRate) {
        int len = std::min(outLen, inLen);
        std::memcpy(to, from, len * sizeof(*from));
        fOutBufferEnd += len;
        fInBufferPos += len;
    } else {
        q->doResampling(to, from, outLen, inLen);
        //cerr << "resampled: " << outLen << '\n';
        fOutBufferEnd += outLen;
        fInBufferPos += inLen;
    }
    if (fInBufferPos >= fInBufferEnd) {
        fInBufferPos = fInBufferEnd = 0;
    }
    return outLen;
}


AudioResampler::AudioResampler()
    : d(new AudioResampler_priv(this))
{
}


AudioResampler::~AudioResampler()
{
    delete d;
}


void
AudioResampler::setDecoder(AudioDecoder* decoder)
{
    d->fDecoder = decoder;
}


int
AudioResampler::setSpec(unsigned dstRate, unsigned channels, unsigned chunkSize)
{
    d->fDstRate = dstRate;
    d->fChannels = channels;
    d->fChunkSize = chunkSize;
    unsigned srcChannels;
    d->fDecoder->getSpec(srcChannels, d->fSrcRate);
    d->fSrcRate = std::min(std::max(4000U, d->fSrcRate), 192000U);
    d->fAdjustBufferSizes();
    adjustForOutputSpec(d->fDstRate, d->fSrcRate, d->fChannels);
    return 0;
}


unsigned
AudioResampler::currentRate() const
{
    return d->fDstRate;
}


unsigned
AudioResampler::currentChannels() const
{
    return d->fChannels;
}


unsigned
AudioResampler::currentChunkSize() const
{
    return d->fChunkSize;
}


int
AudioResampler::resample(float dst[], int dstLen)
{
    //cerr << '\n';
    //std::cerr << dstLen << '\n';
    //return 0;

    int totalSamples = 0;
    int wantedSamples = dstLen;
    bool decEOF = false;

    if (d->fPendingSpecChange) {
        totalSamples += d->fMoveFromOutBuffer(dst, dstLen);
        d->fRelocateOutBuffer();
        d->fResampleFromInBuffer();
        if (totalSamples >= wantedSamples) {
            return totalSamples;
        }
        setSpec(d->fDstRate, d->fChannels, d->fChunkSize);
        d->fPendingSpecChange = false;
    }

    while (totalSamples < wantedSamples and not decEOF) {
        if (d->fInBufferEnd < d->fInBufferSize) {
            bool callAgain = false;
            //AM_debugPrintLn("requesting: " << d->fInBufferSize - d->fInBufferEnd);
            int decSamples = d->fDecoder->decode(d->fInBuffer + d->fInBufferEnd,
                                                 d->fInBufferSize - d->fInBufferEnd,
                                                 callAgain);
            //cerr << "decoded " << decSamples << " samples to inbuf at pos " << fInBufferEnd << endl;

            if (callAgain) {
                d->fInBufferEnd += decSamples;
                d->fResampleFromInBuffer();
                totalSamples += d->fMoveFromOutBuffer(dst + totalSamples,
                                                      dstLen - totalSamples);
                //cerr << "total: " << totalSamples << " wanted: " << wantedSamples << endl;
                if (totalSamples >= wantedSamples) {
                    d->fPendingSpecChange = true;
                    return totalSamples;
                }
                setSpec(d->fDstRate, d->fChannels, d->fChunkSize);
            } else if (decSamples <= 0) {
                decEOF = true;
            } else {
                d->fInBufferEnd += decSamples;
            }
        }

        d->fResampleFromInBuffer();
        d->fRelocateInBuffer();
        totalSamples += d->fMoveFromOutBuffer(dst + totalSamples,
                                              dstLen - totalSamples);
        d->fRelocateOutBuffer();

        if (d->fInBufferSize - d->fInBufferEnd == 0) {
            (void)5;
        }
    }
    //cerr << "total: " << totalSamples << endl;
    return totalSamples;
}
