/*
  SDL_Audiolib - An audio mixing, decoding and resampling library
  Copyright (C) 2013  Nikos Chantziaras

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/
#include "audiostream.h"
#include "am_util.h"

//#include <iostream>
#include <vector>
#include <algorithm>
#include <limits>
#include <cmath>
#include <cstdlib>
#include <SDL.h>
#include <SDL_audio.h>
#include <SDL_config.h>
#include <SDL_version.h>
#include "audiodecoder.h"
#include "dec_vorbis.h"
#include "dec_mpg123.h"
#include "dec_modplug.h"
#include "dec_fluidsynth.h"
#include "resamp_speex.h"
#include "resamp_src.h"

using Aulib::AudioStream;


// This points to an appropriate converter for the current audio format.
static void (*sampleConverter)(Uint8[], float[], int) = nullptr;


// The SDL audio callback uses C calling conventions, so we need to call our
// real callback implementation (which is a member function) through a pointer.
static void(*sdlCallbackImpl)(void*, Uint8[], int) = nullptr;


/* Convert and clip a float sample to an integer sample. This works for
 * all supported integer sample types (8-bit, 16-bit, 32-bit, signed or
 * unsigned.)
 */
template <typename T>
static void
floatSampleToInt(T& dst, float src)
{
    if (src >= 1.f) {
        // Overflow. Clip to max.
        //std::cerr << "+clipping\n";
        dst = std::numeric_limits<T>::max();
    } else if (src < -1.f) {
        // Underflow. Clip to min.
        //std::cerr << "-clipping\n";
        dst = std::numeric_limits<T>::min();
    } else {
        dst = src * (float)(1UL << (sizeof(T) * 8 - 1))
              + ((float)(1UL << (sizeof(T) * 8 - 1))
                 + (float)std::numeric_limits<T>::min());
    }
}


template <typename T>
static void
floatBufToInt(T dst[], const float src[], int srcLen)
{
    for (int i = 0; i < srcLen; ++i) {
        floatSampleToInt(dst[i], src[i]);
    }
}


static void
floatToS8(Uint8 dst[], float src[], int srcLen)
{
    floatBufToInt((Sint8*)dst, src, srcLen);
}


static void
floatToU8(Uint8 dst[], float src[], int srcLen)
{
    floatBufToInt(dst, src, srcLen);
}


static void
floatToS16LSB(Uint8 dst[], float src[], int srcLen)
{
#if SDL_BYTEORDER == SDL_LIL_ENDIAN
    floatBufToInt((Sint16*)dst, src, srcLen);
#else
    Sint16 sample;
    for (int i = 0; i < srcLen; ++i) {
        floatSampleToInt(sample, src[i]);
        *dst++ = *(Uint8*)((unsigned char*)&sample + 1);
        *dst++ = *(Uint8*)((unsigned char*)&sample);
    }
#endif
}


static void
floatToU16LSB(Uint8 dst[], float src[], int srcLen)
{
#if SDL_BYTEORDER == SDL_LIL_ENDIAN
    floatBufToInt((Uint16*)dst, src, srcLen);
#else
    Uint16 sample;
    for (int i = 0; i < srcLen; ++i) {
        floatSampleToInt(sample, src[i]);
        *dst++ = *(Uint8*)((unsigned char*)&sample + 1);
        *dst++ = *(Uint8*)((unsigned char*)&sample);
    }
#endif
}


static void
floatToS16MSB(Uint8 dst[], float src[], int srcLen)
{
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
    floatBufToInt((Sint16*)dst, src, srcLen);
#else
    Sint16 sample;
    for (int i = 0; i < srcLen; ++i) {
        floatSampleToInt(sample, src[i]);
        *dst++ = *(Uint8*)((unsigned char*)&sample + 1);
        *dst++ = *(Uint8*)((unsigned char*)&sample);
    }
#endif
}


static void
floatToU16MSB(Uint8 dst[], float src[], int srcLen)
{
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
    floatBufToInt((Uint16*)dst, src, srcLen);
#else
    Uint16 sample;
    for (int i = 0; i < srcLen; ++i) {
        floatSampleToInt(sample, src[i]);
        *dst++ = *(Uint8*)((unsigned char*)&sample + 1);
        *dst++ = *(Uint8*)((unsigned char*)&sample);
    }
#endif
}


#if SDL_VERSION_ATLEAST(1,3,0)
static void
floatToS32LSB(Uint8 dst[], float src[], int srcLen)
{
#if SDL_BYTEORDER == SDL_LIL_ENDIAN
    floatBufToInt((Sint32*)dst, src, srcLen);
#else
    // FIXME: 4 bytes
    abort();
    Sint32 sample;
    for (int i = 0; i < srcLen; ++i) {
        floatSampleToInt(sample, src[i]);
        *dst++ = *(Uint8*)((unsigned char*)&sample + 1);
        *dst++ = *(Uint8*)((unsigned char*)&sample);
    }
#endif
}


static void
floatToFloat(Uint8 dst[], float src[], int srcLen)
{
    memcpy(dst, src, srcLen * sizeof(*src));
}
#endif


extern "C" {
static void
sdlCallback(void*, Uint8 out[], int outLen)
{
    AM_debugAssert(::sdlCallbackImpl != nullptr);
    ::sdlCallbackImpl(nullptr, out, outLen);
}
}


class Aulib::AudioStream_priv {
    friend class AudioStream;
    AudioStream* q;

    AudioStream_priv(AudioStream* pub, AudioDecoder *decoder, AudioResampler *resampler);
    ~AudioStream_priv();

    bool fIsOpen;
    SDL_RWops* fRWops;
    class AudioDecoder* fDecoder;
    class AudioResampler* fResampler;
    float* fPreBuffer;
    int fPreBufferSize;
    bool fIsPlaying;
    float fVolume;
    unsigned fCurrentIteration;
    unsigned fWantedIterations;

    static SDL_AudioSpec fAudioSpec;
    static std::vector<AudioStream*> fStreamList;

    static void fSdlCallbackImpl(void*, Uint8 out[], int outLen);
};


SDL_AudioSpec Aulib::AudioStream_priv::fAudioSpec;
std::vector<AudioStream*> Aulib::AudioStream_priv::fStreamList;


Aulib::AudioStream_priv::AudioStream_priv(AudioStream* pub, AudioDecoder* decoder,
                                          AudioResampler* resampler)
    : q(pub)
{
    fIsOpen = false;
    fDecoder = decoder;
    fResampler = resampler;
    fPreBufferSize = 0;
    fIsPlaying = false;
    fVolume = 1.f;
    fCurrentIteration = 0;
    fWantedIterations = 0;
    if (resampler) {
        resampler->setDecoder(fDecoder);
    }
}


Aulib::AudioStream_priv::~AudioStream_priv()
{
    q->stop();
    delete fResampler;
    delete fDecoder;
    if (fRWops) {
        SDL_FreeRW(fRWops);
    }
}


void
Aulib::AudioStream_priv::fSdlCallbackImpl(void*, Uint8 out[], int outLen)
{
    AM_debugAssert(sampleConverter != nullptr);

    int wantedSamples = outLen / ((fAudioSpec.format & 0xFF) / 8);

    //cerr << "callback len: " << outLen << '\n';

    static int bufSize = wantedSamples;
    static float* finalMixBuffer = new float[bufSize];
    static float* strmBuf = new float[bufSize];
    if (bufSize != wantedSamples) {
        bufSize = wantedSamples;
        delete[] finalMixBuffer;
        delete[] strmBuf;
        finalMixBuffer = new float[bufSize];
        strmBuf = new float[bufSize];
    }
    std::fill_n(finalMixBuffer, bufSize, 0.f);

    // Iterate over a copy of the original stream list, since we might want to
    // modify the original as we go, removing streams that have stopped.
    std::vector<AudioStream*> streamList(fStreamList);

    for (size_t i = 0; i < streamList.size(); ++i) {
        AudioStream& stream = *streamList[i];
        if (stream.d->fWantedIterations != 0
            and stream.d->fCurrentIteration >= stream.d->fWantedIterations)
        {
            continue;
        }
        int len = 0;
        while (len < wantedSamples) {
            if (stream.d->fResampler) {
                len += stream.d->fResampler->resample(strmBuf + len, wantedSamples - len);
            } else {
                bool callAgain = true;
                while (len < wantedSamples and callAgain) {
                    len += stream.d->fDecoder->decode(strmBuf + len,
                                                      wantedSamples - len,
                                                      callAgain);
                }
            }
            if (len < wantedSamples) {
                stream.d->fDecoder->rewind();
                if (stream.d->fWantedIterations != 0) {
                    ++stream.d->fCurrentIteration;
                    if (stream.d->fCurrentIteration >= stream.d->fWantedIterations) {
                        stream.d->fIsPlaying = false;
                        fStreamList.erase(std::remove(fStreamList.begin(), fStreamList.end(), &stream),
                                             fStreamList.end());
                        stream.invokeFinishCallback();
                        break;
                    }
                    stream.invokeLoopCallback();
                }
            }
        }
        // Avoid mixing when volume is <= 0.
        if (stream.d->fVolume > 0.f) {
            // Avoid scaling operation when volume is 1.
            if (stream.d->fVolume != 1.f) {
                for (int i = 0; i < len; ++i) {
                    finalMixBuffer[i] += strmBuf[i] * stream.d->fVolume;
                }
            } else {
                for (int i = 0; i < len; ++i) {
                    finalMixBuffer[i] += strmBuf[i];
                }
            }
        }
    }
    sampleConverter(out, finalMixBuffer, wantedSamples);
}


AudioStream::AudioStream(const char* filename, AudioDecoder* decoder,
                         AudioResampler* resampler)
    : d(new AudioStream_priv(this, decoder, resampler))
{
    d->fRWops = SDL_RWFromFile(filename, "rb");
}


AudioStream::AudioStream(SDL_RWops* rwops, AudioDecoder* decoder,
                         AudioResampler* resampler)
    : d(new AudioStream_priv(this, decoder, resampler))
{
    d->fRWops = rwops;
}


AudioStream::~AudioStream()
{
    delete d;
}


int
AudioStream::open()
{
    if (d->fIsOpen) {
        return 0;
    }
    if (d->fRWops == nullptr) {
        return -1;
    }
    if (d->fDecoder->open(d->fRWops) < 0) {
        return -1;
    }
    if (d->fResampler) {
        d->fResampler->setSpec(spec().freq, spec().channels, spec().samples);
    }
    d->fIsOpen = true;
    return 0;
}


int
AudioStream::play(unsigned iterations)
{
    if (not d->fIsOpen and open() < 0) {
        return -1;
    }
    if (d->fIsPlaying) {
        return 0;
    }
    d->fCurrentIteration = 0;
    d->fWantedIterations = iterations;
    SDL_LockAudio();
    d->fStreamList.push_back(this);
    d->fIsPlaying = true;
    SDL_UnlockAudio();
    return 0;
}


int
AudioStream::stop()
{
    if (not d->fIsOpen and open() < 0) {
        return -1;
    }
    if (not d->fIsPlaying) {
        return 0;
    }
    SDL_LockAudio();
    d->fIsPlaying = false;
    SDL_UnlockAudio();
    return 0;
}


int
AudioStream::rewind()
{
    if (not d->fIsOpen and open() < 0) {
        return -1;
    }
    SDL_LockAudio();
    int ret = d->fDecoder->rewind();
    SDL_UnlockAudio();
    return ret;
}


int
AudioStream::setVolume(float volume)
{
    if (volume < 0.f) {
        volume = 0.f;
    }
    d->fVolume = volume;
    return 0;
}


float
AudioStream::volume() const
{
    return d->fVolume;
}


bool
AudioStream::isPlaying() const
{
    return d->fIsPlaying;
}


Sint64 AudioStream::lengthMs() const
{
    return d->fDecoder->lengthMs();
}


int
AudioStream::init(int freq, int format, int channels, int bufferSize)
{
    if (SDL_InitSubSystem(SDL_INIT_AUDIO) != 0) {
        return -1;
    }
    SDL_AudioSpec requestedSpec;
    memset(&requestedSpec, 0, sizeof(SDL_AudioSpec));
    requestedSpec.freq = freq;
    requestedSpec.format = format;
    requestedSpec.channels = channels;
    requestedSpec.samples = bufferSize;
    requestedSpec.userdata = nullptr;
    requestedSpec.callback = ::sdlCallback;
    AudioStream_priv::fAudioSpec = requestedSpec;
    if (SDL_OpenAudio(&requestedSpec, &AudioStream_priv::fAudioSpec) != 0) {
    //if (SDL_OpenAudio(&AudioStream_priv::fAudioSpec, nullptr) < 0) {
        AudioStream::quit();
        return -1;
    }

    AM_debugPrint("format: ");
    switch (AudioStream_priv::fAudioSpec.format) {
        case AUDIO_S8:
            AM_debugPrintLn("S8");
            sampleConverter = floatToS8;
            break;
        case AUDIO_U8:
            AM_debugPrintLn("U8");
            sampleConverter = floatToU8;
            break;
        case AUDIO_S16LSB:
            AM_debugPrintLn("S16LSB");
            sampleConverter = floatToS16LSB;
            break;
        case AUDIO_U16LSB:
            AM_debugPrintLn("U16LSB");
            sampleConverter = floatToU16LSB;
            break;
        case AUDIO_S16MSB:
            AM_debugPrintLn("S16MSB");
            sampleConverter = floatToS16MSB;
            break;
        case AUDIO_U16MSB:
            AM_debugPrintLn("U16MSB");
            sampleConverter = floatToU16MSB;
            break;
#if SDL_VERSION_ATLEAST(1,3,0)
        case AUDIO_S32LSB:
            AM_debugPrintLn("S32LSB");
            sampleConverter = floatToS32LSB;
            break;
        case AUDIO_F32LSB:
            AM_debugPrintLn("F32LSB");
            sampleConverter = floatToFloat;
            break;
#endif
        default:
            AM_debugPrintLn("Unknown audio format spec: " << AudioStream_priv::fAudioSpec.format);
            AudioStream::quit();
            //return -1;
    }
    ::sdlCallbackImpl = AudioStream_priv::fSdlCallbackImpl;

#if 1
    AM_debugPrintLn("spec.samples: " << AudioStream_priv::fAudioSpec.samples << '\n'
                    << "spec.freq: "  << AudioStream_priv::fAudioSpec.freq << '\n'
                    << "spec.channels: " << (int)AudioStream_priv::fAudioSpec.channels << '\n'
                    << "spec.silence: " << (int)AudioStream_priv::fAudioSpec.silence << '\n'
                    << "spec.size: " << AudioStream_priv::fAudioSpec.size << "\n");
#endif

    SDL_PauseAudio(false);
    return 0;
}


void
AudioStream::quit()
{
    for (size_t i = 0; i < AudioStream_priv::fStreamList.size(); ++i) {
        AudioStream_priv::fStreamList[i]->stop();
    }
    SDL_QuitSubSystem(SDL_INIT_AUDIO);
    sampleConverter = nullptr;
}


const SDL_AudioSpec&
AudioStream::spec()
{
    return AudioStream_priv::fAudioSpec;
}
