/*
  SDL_Audiolib - An audio mixing, decoding and resampling library
  Copyright (C) 2013  Nikos Chantziaras

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/
#include "dec_flac.h"

#include <iostream>
#include <FLAC/stream_decoder.h>
#include <SDL_version.h>
#include <SDL_rwops.h>


using Aulib::AudioDecoderFlac;
using Aulib::AudioDecoderFlac_priv;


struct Aulib::AudioDecoderFlac_priv {
    AudioDecoderFlac_priv();
    ~AudioDecoderFlac_priv();

    SDL_RWops* fRwops;
    FLAC__StreamDecoder* fFlacDecoder;
    bool fEOF;
    Sint64 fLenMs;
    float* fDstBuf;
    int fDstLen;
};


/* FLAC decoder read callback.
 */
static FLAC__StreamDecoderReadStatus
flacReadCb(const FLAC__StreamDecoder*, FLAC__byte dst[],
           size_t* bytes, void* priv)
{
    AudioDecoderFlac_priv* d = static_cast<AudioDecoderFlac_priv*>(priv);
    int bytesRead = SDL_RWread(d->fRwops, dst, 1, *bytes);
    if (bytesRead <= 0) {
        *bytes = 0;
        return FLAC__STREAM_DECODER_READ_STATUS_END_OF_STREAM;
    }
    *bytes = bytesRead;
    return FLAC__STREAM_DECODER_READ_STATUS_CONTINUE;
}


/* FLAC decoder seek callback.
 */
static FLAC__StreamDecoderSeekStatus
flacSeekCb(const FLAC__StreamDecoder*, FLAC__uint64 offset, void* priv)
{
    AudioDecoderFlac_priv* d = static_cast<AudioDecoderFlac_priv*>(priv);
    if (SDL_RWseek(d->fRwops, offset, RW_SEEK_SET) < 0) {
        return FLAC__STREAM_DECODER_SEEK_STATUS_ERROR;
    }
    return FLAC__STREAM_DECODER_SEEK_STATUS_OK;
}


/* FLAC decoder tell callback.
 */
static FLAC__StreamDecoderTellStatus
flacTellCb(const FLAC__StreamDecoder*, FLAC__uint64* offset, void* priv)
{
    AudioDecoderFlac_priv* d = static_cast<AudioDecoderFlac_priv*>(priv);
    int pos = SDL_RWtell(d->fRwops);
    if (pos < 0) {
        return FLAC__STREAM_DECODER_TELL_STATUS_ERROR;
    }
    *offset = pos;
    return FLAC__STREAM_DECODER_TELL_STATUS_OK;
}


/* FLAC decoder length callback.
 */
static FLAC__StreamDecoderLengthStatus
flacLengthCb(const FLAC__StreamDecoder*, FLAC__uint64* len, void* priv)
{
    AudioDecoderFlac_priv* d = static_cast<AudioDecoderFlac_priv*>(priv);
#if SDL_VERSION_ATLEAST(1,3,0)
    Sint64 size = SDL_RWsize(d->fRwops);
    if (size < 0) {
        return FLAC__STREAM_DECODER_LENGTH_STATUS_ERROR;
    }
    *len = size;
    return FLAC__STREAM_DECODER_LENGTH_STATUS_OK;
#else
    int curPos = SDL_RWtell(d->fRwops);
    SDL_RWseek(d->fRwops, 0, RW_SEEK_END);
    int endPos = SDL_RWtell(d->fRwops);
    SDL_RWseek(d->fRwops, curPos, RW_SEEK_SET);
    if (endPos < 0) {
        return FLAC__STREAM_DECODER_LENGTH_STATUS_ERROR;
    }
    *len = endPos;
    return FLAC__STREAM_DECODER_LENGTH_STATUS_OK;
#endif
}


/* FLAC decoder EOF callback.
 */
static FLAC__bool
flacEofCb(const FLAC__StreamDecoder*, void*)
{
    return false;
}


/* FLAC decoder write callback.
 */
static FLAC__StreamDecoderWriteStatus
flacWriteCb(const FLAC__StreamDecoder*, const FLAC__Frame* frame,
            const FLAC__int32* const buffer[], void* priv)
{
}


/* FLAC decoder error callback.
 */
static void
flacErrorCb(const FLAC__StreamDecoder*, FLAC__StreamDecoderErrorStatus, void*)
{
}


AudioDecoderFlac_priv::AudioDecoderFlac_priv()
    : fRwops(nullptr),
      fFlacDecoder(nullptr),
      fEOF(false),
      fLenMs(-1),
      fDstBuf(nullptr),
      fDstLen(0)
{
}


Aulib::AudioDecoderFlac_priv::~AudioDecoderFlac_priv()
{
}


AudioDecoderFlac::AudioDecoderFlac()
    : d(new AudioDecoderFlac_priv)
{
}


AudioDecoderFlac::~AudioDecoderFlac()
{
    delete d;
}


int
AudioDecoderFlac::open(SDL_RWops* rwops)
{
    if (d->fFlacDecoder) {
        return 0;
    }
    d->fFlacDecoder = FLAC__stream_decoder_new();
    if (d->fFlacDecoder == nullptr) {
        return -1;
    }
    d->fRwops = rwops;
    FLAC__stream_decoder_init_stream(d->fFlacDecoder,
                                     flacReadCb,
                                     rwops->seek ? flacSeekCb : nullptr,
                                     rwops->seek ? flacTellCb : nullptr,
                                     rwops->seek ? flacLengthCb : nullptr,
                                     rwops->seek ? flacEofCb : nullptr,
                                     flacWriteCb,
                                     nullptr,
                                     flacErrorCb,
                                     this->d);
    return 0;
}


void
AudioDecoderFlac::getSpec(unsigned& channels, unsigned& rate)
{
}


int
AudioDecoderFlac::decode(float buf[], int len, bool& callAgain)
{
    return -1;
}


int
AudioDecoderFlac::rewind()
{
    return -1;
}


Sint64
AudioDecoderFlac::lengthMs() const
{
    return -1;
}
