/*
  SDL_Audiolib - An audio mixing, decoding and resampling library
  Copyright (C) 2013  Nikos Chantziaras

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/
#include "dec_fluidsynth.h"

#include <iostream>
#include <SDL_rwops.h>
#include <SDL_audio.h>
#include <fluidsynth.h>
#include "audiostream.h"

using Aulib::AudioDecoderFluidSynth;


static fluid_settings_t* settings = nullptr;


static int
initFluidSynth()
{
    if (settings) {
        return 0;
    }
    if ((settings = new_fluid_settings()) == nullptr) {
        return -1;
    }
    fluid_settings_setnum(settings, "synth.sample-rate",
                          Aulib::AudioStream::spec().freq);
    return 0;
}


class Aulib::AudioDecoderFluidSynth_priv {
    friend class AudioDecoderFluidSynth;

    AudioDecoderFluidSynth_priv();
    ~AudioDecoderFluidSynth_priv();

    fluid_synth_t* fSynth;
    fluid_player_t* fPlayer;
    Uint8* fMidiData;
    int fMidiDataLen;
    bool fEOF;
};


Aulib::AudioDecoderFluidSynth_priv::AudioDecoderFluidSynth_priv()
    : fPlayer(nullptr),
      fMidiData(nullptr),
      fMidiDataLen(0),
      fEOF(false)
{
    if (settings == nullptr) {
        initFluidSynth();
    }
    if ((fSynth = new_fluid_synth(settings)) == nullptr) {
        return;
    }
    fluid_synth_set_gain(fSynth, 0.7f);
    fluid_synth_set_interp_method(fSynth, -1, FLUID_INTERP_7THORDER);
}


Aulib::AudioDecoderFluidSynth_priv::~AudioDecoderFluidSynth_priv()
{
    delete_fluid_player(fPlayer);
    delete_fluid_synth(fSynth);
    delete[] fMidiData;
}


AudioDecoderFluidSynth::AudioDecoderFluidSynth()
    : d(new AudioDecoderFluidSynth_priv)
{
}


AudioDecoderFluidSynth::~AudioDecoderFluidSynth()
{
    delete d;
}


int
AudioDecoderFluidSynth::loadSoundfont(const char filename[])
{
    fluid_synth_sfload(d->fSynth, filename, true);
    return 0;
}


int
AudioDecoderFluidSynth::open(SDL_RWops* rwops)
{
    if (d->fSynth == nullptr)
        return -1;

    int frontPos = SDL_RWtell(rwops);
    d->fMidiDataLen = SDL_RWseek(rwops, 0, RW_SEEK_END) - frontPos;
    SDL_RWseek(rwops, frontPos, RW_SEEK_SET);
    if (d->fMidiDataLen <= 0) {
        return -1;
    }

    d->fMidiData = new Uint8[d->fMidiDataLen];
    if (SDL_RWread(rwops, d->fMidiData, d->fMidiDataLen, 1) != 1
        or (d->fPlayer = new_fluid_player(d->fSynth)) == nullptr
        or fluid_player_add_mem(d->fPlayer, d->fMidiData, d->fMidiDataLen) != FLUID_OK
        or fluid_player_play(d->fPlayer) != FLUID_OK)
    {
        if (d->fPlayer) {
            delete_fluid_player(d->fPlayer);
        }
        if (d->fSynth) {
            delete_fluid_synth(d->fSynth);
            d->fSynth = nullptr;
        }
        delete[] d->fMidiData;
        return -1;
    }
    return 0;
}


void
AudioDecoderFluidSynth::getSpec(unsigned& channels, unsigned& rate)
{
    int chanRet;
    double rateRet;
    fluid_settings_getint(settings, "synth.audio-channels", &chanRet);
    fluid_settings_getnum(settings, "synth.sample-rate", &rateRet);
    channels = chanRet;
    rate = rateRet;
}


int
AudioDecoderFluidSynth::decode(float buf[], int len, bool& callAgain)
{
    callAgain = false;
    if (d->fSynth == nullptr or d->fEOF) {
        return 0;
    }

    len /= AudioStream::spec().channels;
    int res = fluid_synth_write_float(d->fSynth, len, buf, 0, 2, buf, 1, 2);
    if (fluid_player_get_status(d->fPlayer) == FLUID_PLAYER_DONE) {
        d->fEOF = true;
    }
    if (res == FLUID_OK) {
        return len * AudioStream::spec().channels;
    }
    return 0;
}


int
AudioDecoderFluidSynth::rewind()
{
    fluid_player_stop(d->fPlayer);
    delete_fluid_player(d->fPlayer);
    d->fPlayer = new_fluid_player(d->fSynth);
    fluid_player_add_mem(d->fPlayer, d->fMidiData, d->fMidiDataLen);
    fluid_player_play(d->fPlayer);
    d->fEOF = false;
    return 0;
}


Sint64
AudioDecoderFluidSynth::lengthMs() const
{
    return -1;
}
