/*
  SDL_Audiolib - An audio mixing, decoding and resampling library
  Copyright (C) 2013  Nikos Chantziaras

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/
#include "dec_modplug.h"
#include "am_util.h"

#include <iostream>
#include <libmodplug/modplug.h>
#include <SDL_audio.h>
#include "audiostream.h"

using Aulib::AudioDecoderModPlug;

static ModPlug_Settings* modplugSettings = nullptr;


static void
initModPlug(const SDL_AudioSpec& spec)
{
    modplugSettings = new ModPlug_Settings;
    ModPlug_GetSettings(modplugSettings);
    modplugSettings->mFlags =   MODPLUG_ENABLE_OVERSAMPLING
                              | MODPLUG_ENABLE_NOISE_REDUCTION;
    // TODO: More than 2 channels.
    modplugSettings->mChannels = spec.channels == 1 ? 1 : 2;
    // It seems MogPlug does resample to any samplerate. 32, 44.1, up to
    // 192K all seem to work correctly.
    modplugSettings->mFrequency = spec.freq;
    modplugSettings->mResamplingMode = MODPLUG_RESAMPLE_FIR;
    modplugSettings->mBits = 32;
    ModPlug_SetSettings(modplugSettings);
}


class Aulib::AudioDecoderModPlug_priv {
    friend class AudioDecoderModPlug;

    AudioDecoderModPlug_priv();
    ~AudioDecoderModPlug_priv();

    ModPlugFile* mpHandle;
    bool atEOF;
    Sint64 fLenMs;
};


Aulib::AudioDecoderModPlug_priv::AudioDecoderModPlug_priv()
    : mpHandle(nullptr),
      atEOF(false),
      fLenMs(-1)
{
    if (modplugSettings == nullptr)
        initModPlug(AudioStream::spec());
}


Aulib::AudioDecoderModPlug_priv::~AudioDecoderModPlug_priv()
{
    if (mpHandle)
        ModPlug_Unload(mpHandle);
}


AudioDecoderModPlug::AudioDecoderModPlug()
    : d(new AudioDecoderModPlug_priv)
{
}


AudioDecoderModPlug::~AudioDecoderModPlug()
{
    delete d;
}


int
AudioDecoderModPlug::open(SDL_RWops* rwops)
{
    if (d->mpHandle) {
        return 0;
    }
    int frontPos = SDL_RWtell(rwops);
    int dataSize = SDL_RWseek(rwops, 0, RW_SEEK_END) - frontPos;
    SDL_RWseek(rwops, frontPos, RW_SEEK_SET);
    if (dataSize <= 0) {
        return -1;
    }
    int ret = 0;
    Uint8* data = new Uint8[dataSize];
    if (SDL_RWread(rwops, data, dataSize, 1) != 1) {
        ret = -1;
    } else if ((d->mpHandle = ModPlug_Load(data, dataSize)) == nullptr) {
        ret = -1;
    }
    //ModPlug_SetMasterVolume(d->mpHandle, 192);
    d->fLenMs = ModPlug_GetLength(d->mpHandle);
    delete[] data;
    return ret;
}


void
AudioDecoderModPlug::getSpec(unsigned& channels, unsigned& rate)
{
    channels = modplugSettings->mChannels;
    rate = modplugSettings->mFrequency;
}


int
AudioDecoderModPlug::decode(float buf[], int len, bool& callAgain)
{
    callAgain = false;
    if (d->atEOF) {
        return 0;
    }
    Sint32* data = new Sint32[len];
    int ret = ModPlug_Read(d->mpHandle, data, len * 4);
    for (int i = 0; i < len; ++i) {
        buf[i] = (float)data[i] / 2147483648.f;
    }
    if (ret == 0) {
        d->atEOF = true;
    }
    delete[] data;
    return ret / sizeof(*buf);
}


int
AudioDecoderModPlug::rewind()
{
    ModPlug_Seek(d->mpHandle, 0);
    d->atEOF = false;
    return 0;
}


Sint64
AudioDecoderModPlug::lengthMs() const
{
    return d->fLenMs;
}
