/*
  SDL_Audiolib - An audio mixing, decoding and resampling library
  Copyright (C) 2013  Nikos Chantziaras

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/
#include "dec_mpg123.h"

#include <iostream>
#include <unistd.h>
#include <mpg123.h>
#include <SDL_rwops.h>
#include "am_util.h"

using Aulib::AudioDecoderMpg123;
using Aulib::AudioDecoderMpg123_priv;


static bool initialized = false;


static int
initLibMpg()
{
    if (mpg123_init() != MPG123_OK)
        return -1;
    initialized = true;
    return 0;
}


static int
initMpgFormats(mpg123_handle* handle)
{
    const long* list;
    size_t len;
    mpg123_rates(&list, &len);
    mpg123_format_none(handle);
    for (size_t i = 0; i < len; ++i) {
        if (mpg123_format(handle, list[i], MPG123_STEREO | MPG123_MONO,
                          MPG123_ENC_FLOAT_32) != MPG123_OK)
        {
            return -1;
        }
    }
    return 0;
}


static ssize_t
mpgReadCallback(void* rwops, void* buf, size_t len)
{
    return SDL_RWread(static_cast<SDL_RWops*>(rwops), buf, 1, len);
}


static off_t
mpgSeekCallback(void* rwops, off_t pos, int whence)
{
    switch (whence) {
        case SEEK_SET:
            whence = RW_SEEK_SET;
            break;
        case SEEK_CUR:
            whence = RW_SEEK_CUR;
            break;
        default:
            whence = RW_SEEK_END;
    }
    return SDL_RWseek(static_cast<SDL_RWops*>(rwops), pos, whence);
}


class Aulib::AudioDecoderMpg123_priv {
    friend class AudioDecoderMpg123;

    AudioDecoderMpg123_priv();
    ~AudioDecoderMpg123_priv();

    mpg123_handle* fMpgHandle;
    unsigned fChannels;
    unsigned fRate;
    bool fEOF;
    Sint64 fLenMs;
};


Aulib::AudioDecoderMpg123_priv::AudioDecoderMpg123_priv()
    : fMpgHandle(nullptr),
      fChannels(0),
      fRate(0),
      fEOF(false),
      fLenMs(-1)
{
    if (not initialized)
        initLibMpg();
}


Aulib::AudioDecoderMpg123_priv::~AudioDecoderMpg123_priv()
{
    mpg123_delete(fMpgHandle);
}


AudioDecoderMpg123::AudioDecoderMpg123()
    : d(new AudioDecoderMpg123_priv)
{ }


AudioDecoderMpg123::~AudioDecoderMpg123()
{
    delete d;
}


int
AudioDecoderMpg123::open(SDL_RWops* rwops)
{
    if (d->fMpgHandle) {
        return 0;
    }
    if (not initialized) {
        return -1;
    }
    if ((d->fMpgHandle = mpg123_new(nullptr, nullptr)) == nullptr) {
        return -1;
    }
    if (initMpgFormats(d->fMpgHandle) < 0) {
        return -1;
    }
    mpg123_replace_reader_handle(d->fMpgHandle, mpgReadCallback, mpgSeekCallback, nullptr);
    mpg123_open_handle(d->fMpgHandle, rwops);
    long rate;
    int channels, encoding;
    mpg123_getformat(d->fMpgHandle, &rate, &channels, &encoding);
    d->fChannels = channels;
    d->fRate = rate;
    off_t len = mpg123_length(d->fMpgHandle);
    d->fLenMs = len == MPG123_ERR ? -1 : (double)len / rate * 1000;
    std::cerr << "mpg rate: " << rate << "\n";
    std::cerr << "mpg123 decoder: " << mpg123_current_decoder(d->fMpgHandle) << '\n';
    return 0;
}


void
AudioDecoderMpg123::getSpec(unsigned& channels, unsigned& rate)
{
    channels = d->fChannels;
    rate = d->fRate;
}


int
AudioDecoderMpg123::decode(float buf[], int len, bool& callAgain)
{
    callAgain = false;
    if (d->fEOF) {
        return 0;
    }

    size_t bytesWanted = static_cast<size_t>(len) * sizeof(*buf);
    size_t decBytes = 0;
    size_t totalBytes = 0;

    while (totalBytes < bytesWanted and not callAgain) {
        int ret = mpg123_read(d->fMpgHandle, (unsigned char*)buf, bytesWanted, &decBytes);
        totalBytes += decBytes;
        if (ret == MPG123_NEW_FORMAT) {
            long rate;
            int channels, encoding;
            mpg123_getformat(d->fMpgHandle, &rate, &channels, &encoding);
            std::cerr << "mpg123: format change\n";
            d->fChannels = channels;
            d->fRate = rate;
            callAgain = true;
        } else if (ret == MPG123_DONE) {
            d->fEOF = false;
            break;
        }
    }
    return totalBytes / sizeof(*buf);
}


int
AudioDecoderMpg123::rewind()
{
    if (mpg123_seek(d->fMpgHandle, 0, SEEK_SET) < 0) {
        return -1;
    }
    return 0;
}


Sint64
AudioDecoderMpg123::lengthMs() const
{
    return d->fLenMs;
}
