/*
  SDL_Audiolib - An audio mixing, decoding and resampling library
  Copyright (C) 2013  Nikos Chantziaras

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/
#include "dec_sndfile.h"

#include <iostream>
#include <sndfile.h>
#include <SDL_version.h>
#include <SDL_rwops.h>

using Aulib::AudioDecoderSndfile;
using Aulib::AudioDecoderSndfile_priv;


static sf_count_t
sfLenCb(void* rwops)
{
#if SDL_VERSION_ATLEAST(1,3,0)
    Sint64 size = SDL_RWsize((SDL_RWops*)rwops);
    if (size < 0) {
        return 0;
    }
    return size;
#else
    int curPos = SDL_RWtell((SDL_RWops*)rwops);
    SDL_RWseek((SDL_RWops*)rwops, 0, RW_SEEK_END);
    int endPos = SDL_RWtell((SDL_RWops*)rwops);
    SDL_RWseek((SDL_RWops*)rwops, curPos, RW_SEEK_SET);
    if (endPos < 0) {
        return 0;
    }
    return endPos;
#endif
}


static sf_count_t
sfSeekCb(sf_count_t offset, int whence, void* rwops)
{
    //AM_debugPrintLn("SEEK");
    switch (whence) {
        case SEEK_SET:
            whence = RW_SEEK_SET;
            break;
        case SEEK_CUR:
            whence = RW_SEEK_CUR;
            break;
        default:
            whence = RW_SEEK_END;
    }
    int pos = SDL_RWseek((SDL_RWops*)rwops, offset, whence);
    //AM_debugPrintLn("whence: " << whence << ", seek want: " << offset << ", got: " << pos);
    return pos;
}


static sf_count_t
sfReadCb(void* dst, sf_count_t count, void* rwops)
{
    int ret = SDL_RWread((SDL_RWops*)rwops, dst, 1, count);
    //AM_debugPrintLn(ret);
    return ret;
}


static sf_count_t
sfTellCb(void* rwops)
{
    return SDL_RWtell((SDL_RWops*)rwops);
}


struct Aulib::AudioDecoderSndfile_priv {
    AudioDecoderSndfile_priv();
    ~AudioDecoderSndfile_priv();

    SNDFILE* fSndfile;
    SF_INFO fInfo;
    bool fEOF;
    Sint64 fLenMs;
};


AudioDecoderSndfile_priv::AudioDecoderSndfile_priv()
    : fSndfile(nullptr),
      fEOF(false),
      fLenMs(-1)
{
}


AudioDecoderSndfile_priv::~AudioDecoderSndfile_priv()
{
}


AudioDecoderSndfile::AudioDecoderSndfile()
    : d(new AudioDecoderSndfile_priv)
{ }


AudioDecoderSndfile::~AudioDecoderSndfile()
{
    if (d->fSndfile) {
        sf_close(d->fSndfile);
    }
    delete d;
}


int
AudioDecoderSndfile::open(SDL_RWops* rwops)
{
    if (d->fSndfile) {
        return 0;
    }
    d->fInfo.format = 0;
    SF_VIRTUAL_IO cbs;
    cbs.get_filelen = sfLenCb;
    cbs.seek = sfSeekCb;
    cbs.read = sfReadCb;
    cbs.write = nullptr;
    cbs.tell = sfTellCb;
    d->fSndfile = sf_open_virtual(&cbs, SFM_READ, &d->fInfo, rwops);
    if (d->fSndfile == nullptr) {
        return -1;
    }
    return 0;
}


void
AudioDecoderSndfile::getSpec(unsigned& channels, unsigned& rate)
{
    channels = d->fInfo.channels;
    rate = d->fInfo.samplerate;
}


int
AudioDecoderSndfile::decode(float buf[], int len, bool& callAgain)
{
    callAgain = false;
    if (d->fEOF) {
        return 0;
    }
    sf_count_t ret = sf_read_float(d->fSndfile, buf, len);
    //AM_debugPrintLn("(decode) want: " << len << " got: " << ret);
    if (ret == 0) {
        //cerr << "EOF\n";
        d->fEOF = true;
    }
    return ret;
}


int
AudioDecoderSndfile::rewind()
{
    return sf_seek(d->fSndfile, 0, SEEK_SET);
}


Sint64
AudioDecoderSndfile::lengthMs() const
{
    return (double)d->fInfo.frames / d->fInfo.samplerate * 1000;
}
