/*
  SDL_Audiolib - An audio mixing, decoding and resampling library
  Copyright (C) 2013  Nikos Chantziaras

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
*/
#include "dec_vorbis.h"

#include <iostream>
#include <vorbis/vorbisfile.h>
#include <SDL_rwops.h>

using Aulib::AudioDecoderVorbis;
using Aulib::AudioDecoderVorbis_priv;


static size_t
vorbisReadCb(void* ptr, size_t size, size_t nmemb, void* rwops)
{
    return SDL_RWread(static_cast<SDL_RWops*>(rwops), ptr, size, nmemb);
}


static int
vorbisSeekCb(void* rwops, ogg_int64_t offset, int whence)
{
    if (SDL_RWseek(static_cast<SDL_RWops*>(rwops), offset, whence) < 0) {
        return -1;
    }
    return 0;
}


static long
vorbisTellCb(void* rwops)
{
    return SDL_RWtell(static_cast<SDL_RWops*>(rwops));
}


class Aulib::AudioDecoderVorbis_priv {
    friend class AudioDecoderVorbis;

    AudioDecoderVorbis_priv();
    ~AudioDecoderVorbis_priv();

    OggVorbis_File* fVFHandle;
    int fCurrentSection;
    vorbis_info* fCurrentInfo;
    bool fEOF;
    Sint64 fLenMs;
};


AudioDecoderVorbis_priv::AudioDecoderVorbis_priv()
    : fCurrentSection(0),
      fCurrentInfo(nullptr),
      fEOF(false),
      fLenMs(-1)
{
    fVFHandle = new OggVorbis_File;
}


AudioDecoderVorbis_priv::~AudioDecoderVorbis_priv()
{
    delete fVFHandle;
}


AudioDecoderVorbis::AudioDecoderVorbis()
    : d(new AudioDecoderVorbis_priv)
{ }


AudioDecoderVorbis::~AudioDecoderVorbis()
{
    ov_clear(d->fVFHandle);
    delete d;
}


int
AudioDecoderVorbis::open(SDL_RWops* rwops)
{
    // TODO: Check if already opened.
    ov_callbacks cbs;
    memset(&cbs, 0, sizeof(cbs));
    cbs.read_func = vorbisReadCb;
    cbs.seek_func = vorbisSeekCb;
    cbs.tell_func = vorbisTellCb;
    cbs.close_func = nullptr;
    int ret = ov_open_callbacks(rwops, d->fVFHandle, nullptr, 0, cbs);
    d->fCurrentInfo = ov_info(d->fVFHandle, -1);
    double len = ov_time_total(d->fVFHandle, -1);
    d->fLenMs = len == OV_EINVAL ? -1 : len * 1000;
    return ret;
}


void
AudioDecoderVorbis::getSpec(unsigned& channels, unsigned& rate)
{
    if (d->fCurrentInfo == nullptr) {
        channels = 0;
        rate = 0;
    } else {
        channels = d->fCurrentInfo->channels;
        rate = d->fCurrentInfo->rate;
    }
}


int
AudioDecoderVorbis::decode(float buf[], int len, bool& callAgain)
{
    using namespace std;

    callAgain = false;

    if (d->fEOF)
        return 0;

    float** out;
    int decSamples = 0;

    // FIXME: We only deal with 2 channels right now.
    while (decSamples < len and not callAgain) {
        int lastSection = d->fCurrentSection;
        long ret = ov_read_float(d->fVFHandle, &out, len/2 - decSamples/2,
                                 &d->fCurrentSection);
        if (ret == 0) {
            //cerr << "EOF\n";
            d->fEOF = true;
            break;
        }
        if (ret < 0) {
            cerr << "stream error: ";
            switch (ret) {
                case OV_HOLE: cerr << "OV_HOLE\n"; break;
                case OV_EBADLINK: cerr << "OV_EBADLINK\n"; break;
                case OV_EINVAL: cerr << "OV_EINVAL\n"; break;
                default: cerr << "unknown error: " << ret << '\n';
            }
            break;
        }
        if (d->fCurrentSection != lastSection) {
            cerr << "stream change\n";
            d->fCurrentInfo = ov_info(d->fVFHandle, -1);
            callAgain = true;
        }
        // Copy samples to output buffer in interleaved format.
        for (int i = 0; i < ret; ++i) {
            *buf++ = out[0][i];
            *buf++ = out[1][i];
        }
        decSamples += ret * 2;
    }
    return decSamples;
}


int
AudioDecoderVorbis::rewind()
{
    int ret;
    if (d->fEOF) {
        ret = ov_raw_seek(d->fVFHandle, 0);
        d->fEOF = false;
    } else {
        ret = ov_raw_seek_lap(d->fVFHandle, 0);
    }
    return ret != 0 ? -1 : 0;
}


Sint64
AudioDecoderVorbis::lengthMs() const
{
    return d->fLenMs;
}
